# See the file "LICENSE" for the full license governing this code.

require 'tcell_agent/logger'
require 'tcell_agent/version'
require 'tcell_agent/configuration'

require 'tcell_agent/sensor_events/server_agent'

require 'tcell_agent/policies/policy_types'
require 'tcell_agent/policies/policies_manager'
require 'tcell_agent/policies/policy_polling'
require 'tcell_agent/agent/static_agent'
require 'tcell_agent/agent/route_manager'

require 'tcell_agent/routes/table'

require 'tcell_agent/settings_reporter'
require 'tcell_agent/rust/native_agent'

require 'json'

module TCellAgent
  class Agent
    include TCellAgent::ModuleLoggerAccess

    attr_accessor :route_table,
                  :stop_agent

    def initialize
      @stop_agent = false
      @native_agent = nil
      @route_table = TCellAgent::Routes::RouteTable.new
      @policies_manager = PoliciesManager.new(nil)
    end

    def instrument_auth_frameworks
      if defined?(Devise) && TCellAgent.configuration.should_instrument?('devise')
        module_logger.info('Instrumenting Devise authentication framework')
        require 'tcell_agent/rails/auth/devise'
        require 'tcell_agent/rails/auth/devise_helper'
      end

      if defined?(Authlogic) && TCellAgent.configuration.should_instrument?('authlogic')
        module_logger.info('Instrumenting Authlogic authentication framework')
        require 'tcell_agent/rails/auth/authlogic'
        require 'tcell_agent/rails/auth/authlogic_helper'
      end

      if defined?(Doorkeeper) && TCellAgent.configuration.should_instrument?('doorkeeper') # rubocop:disable Style/GuardClause
        module_logger.info('Instrumenting Doorkeeper authentication framework')
        require 'tcell_agent/rails/auth/doorkeeper'
      end
    end

    def instrument_built_ins
      require 'tcell_agent/instrumentation/cmdi'
      require 'tcell_agent/instrumentation/lfi'

      variant = if RUBY_VERSION.start_with?('3')
                  'ruby_3'
                else
                  'ruby_2'
                end

      if TCellAgent.configuration.should_instrument?('io')
        module_logger.info('Instrumenting Ruby Class: IO')
        require "tcell_agent/instrumentation/monkey_patches/#{variant}/io"
      end

      if TCellAgent.configuration.should_instrument?('file')
        module_logger.info('Instrumenting Ruby Class: File')
        require "tcell_agent/instrumentation/monkey_patches/#{variant}/file"
      end

      if TCellAgent.configuration.should_instrument?('kernel') # rubocop:disable Style/GuardClause
        module_logger.info('Instrumenting Ruby Module: Kernel')
        require "tcell_agent/instrumentation/monkey_patches/#{variant}/kernel"
      end
    end

    def manage_policies
      @policies_manager = PoliciesManager.new(@native_agent)

      result = {}
      unless TCellAgent.configuration.should_start_policy_poll?
        result = @native_agent.poll_new_policies
      end

      policies_and_enablements = result['new_policies_and_enablements'] || {}

      @policies_manager.process_policy_json(
        policies_and_enablements['enablements'],
        policies_and_enablements['policies']
      )

      @policy_polling = PolicyPolling.new(@policies_manager, @native_agent)
    end

    def policies
      @policies_manager.policies
    end

    def queue_sensor_event(event)
      return unless @native_agent

      @native_agent.send_sanitized_events(
        [event]
      )
    rescue StandardError => e
      module_logger.error("Error sending event: (#{e.class}) #{e.message}")
      module_logger.exception(e)
    end

    def report_metrics(request_time, tcell_context)
      @native_agent.report_metrics(
        request_time, tcell_context
      )
    rescue StandardError => e
      module_logger.error("Error reporting metric: (#{e.class}) #{e.message}")
      module_logger.exception(e)
    end

    def start(server_name)
      @native_agent = TCellAgent::Rust::NativeAgent.create_agent(
        TCellAgent.initializer_configuration ||
        TCellAgent.configuration
      )

      if @native_agent.nil?
        TCellAgent.configuration.enabled = false
        return
      end

      TCellAgent.native_logger = @native_agent

      module_logger.info('Rails initializer overriding default agent configuration') unless TCellAgent.initializer_configuration.nil?

      instrument_auth_frameworks
      instrument_built_ins
      manage_policies

      module_logger.info("Started thread agent: #{server_name}")
      TCellAgent.report_settings
      TCellAgent::Instrumentation::Rails.send_settings
    rescue StandardError => e
      TCellAgent.configuration.enabled = false
      module_logger.error("Error starting agent: (#{e.class}) #{e.message}")
      module_logger.exception(e)
    end
  end
end
