# frozen_string_literal: true

# See the file "LICENSE" for the full license governing this code.
require 'fileutils'
require 'json'
require 'yaml'
require 'socket'
require 'securerandom'
require 'uri'

module TCellAgent
  class ConfigurationException < StandardError
  end

  class << self
    attr_accessor :configuration
    attr_accessor :initializer_configuration
  end

  def self.configure
    require 'tcell_agent/config_initializer'
    self.initializer_configuration ||= ConfigInitializer.new

    yield(initializer_configuration)
  rescue NoMethodError => e
    logger = TCellAgent::ModuleLogger.new(TCellAgent::RubyLogger.new, name)
    logger.error("Error configuring tcell_agent with initializers: #{e}")
  end

  class Configuration
    # internal Ruby configurations
    attr_accessor :disable_all

    # Common config returned by libtcellagent
    attr_accessor :api_key, :app_id, :disabled_instrumentation, :enabled,
                  :enable_intercept_requests, :fetch_policies_from_tcell, :hmac_key,
                  :host_identifier, :instrument, :log_dir, :logging_options,
                  :password_hmac_key, :reverse_proxy, :reverse_proxy_ip_address_header,
                  :tcell_api_url

    # Ruby config returned by libtcellagent
    attr_accessor :enable_policy_polling

    def should_start_policy_poll?
      @enabled && @enable_policy_polling && @fetch_policies_from_tcell # fetch_policies_from_tcel = legacy
    end

    def should_instrument?(func = nil)
      return false unless @enabled && @instrument # never instrument if disabled
      return true if func.nil?                    # always instrument if enabled and nothing given

      !@disabled_instrumentation.include?(func)
    end

    def should_intercept_requests?
      @enabled && @enable_intercept_requests
    end

    def initialize
      # ruby agent defaults
      @disable_all = false
      @logging_options = {}

      check_for_disabled_instrumentation(get_config_file_name)
    end

    def get_config_file_path
      return nil unless ENV['TCELL_AGENT_HOME']
      return nil if ENV['TCELL_AGENT_CONFIG']

      File.join(Dir.getwd, '/config/tcell_agent.config')
    end

    def get_config_file_name
      ENV['TCELL_AGENT_CONFIG'] || File.join(Dir.getwd, '/config/tcell_agent.config')
    end

    def check_for_disabled_instrumentation(filename)
      return unless File.file?(filename)

      begin
        config = JSON.parse(File.open(filename).read)

        if config['version'] == 1
          app_data = config['applications'][0]
          @disable_all = to_bool(app_data.fetch('disable_all', @disable_all))
        end
      rescue StandardError
        # do nothing
      end
    end

    def to_bool(var)
      { 'true' => true, 'false' => false }[var.to_s.downcase]
    end

    def populate_configuration(native_agent_config_response)
      # config
      @enabled = native_agent_config_response['enabled']
      @disabled_instrumentation = Set.new(native_agent_config_response['disabled_instrumentation'])
      @fetch_policies_from_tcell = native_agent_config_response['update_policy']
      @instrument = native_agent_config_response['instrument']

      # app config
      apps = native_agent_config_response['applications'] ? native_agent_config_response['applications']['first'] : {}
      @app_id = apps['app_id']
      @api_key = apps['api_key']
      @hmac_key = apps['hmac_key']
      @password_hmac_key = apps['password_hmac_key']

      # proxy config
      @reverse_proxy = apps['reverse_proxy']
      @reverse_proxy_ip_address_header = apps['reverse_proxy_ip_address_header']

      # endpoint config
      endpoint = native_agent_config_response['endpoint_config'] || {}
      @tcell_api_url = endpoint['api_url']

      # ruby config
      ruby_config = native_agent_config_response['ruby_config'] || {}
      @enable_policy_polling = ruby_config['enable_policy_polling']
      @enable_intercept_requests = !@disabled_instrumentation.include?('intercept_requests')
    end
  end

  TCellAgent.configuration ||= TCellAgent::Configuration.new
end
