# See the file "LICENSE" for the full license governing this code.
require 'tcell_agent/configuration'
require 'tcell_agent/version'
require 'date'
require 'cgi'

module TCellAgent
  module Instrumentation
    TCELL_ID = 'tcell.request_data'.freeze

    class ContextFilter
      attr_accessor :type
      attr_accessor :rule
      attr_accessor :context
      attr_accessor :parameter
      attr_accessor :database
      attr_accessor :schema
      attr_accessor :table
      attr_accessor :field

      DATABASE = 'db'.freeze
      REQUEST = 'request'.freeze

      def for_request(context, parameter, rule)
        self.type = ContextFilter::REQUEST
        self.context = context
        self.parameter = parameter
        self.rule = rule
        self
      end

      def create_hash_value
        "#{type}#{context}#{parameter}#{database}#{schema}#{table}#{field}#{rule}".hash
      end

      def eql?(other)
        hash == other.hash
      end

      def hash
        create_hash_value
      end

      def for_database(database, schema, table, field, rule)
        self.type = ContextFilter::DATABASE
        self.database = database
        self.schema = schema
        self.table = table
        self.field = field
        self.rule = rule
        self
      end

      def for_request(context, parameter, rule) # rubocop:disable Lint/DuplicateMethods
        self.type = ContextFilter::REQUEST
        self.context = context
        self.parameter = parameter
        self.rule = rule
        self
      end
    end

    class TCellData
      attr_accessor :transaction_id, :session_id, :hmac_session_id, :user_id,
                    :password, :route_id, :path, :uri, :fullpath, :context_filters_by_term,
                    :database_filters, :remote_address, :user_agent, :request_method,
                    :path_parameters, :patches_blocking_triggered, :grape_mount_endpoint,
                    :referrer, :csrf_exception_name, :sql_exceptions, :database_result_sizes,
                    :reverse_proxy_header_value

      def self.filterx(sanitize_string, event_flag, replace_flag, term)
        send_event = false
        sanitize_string.gsub!(term) do |m|
          if replace_flag
            m = '[redacted]'
            send_event = true
          elsif event_flag
            # m = "[hash]"
            send_event = true
          end
          m
        end
        send_event
      end

      def initialize
        @patches_blocking_triggered = false
        @context_filters_by_term = Hash.new { |h, k| h[k] = Set.new }
        @sql_exceptions = []
        @database_result_sizes = []
      end

      def valid_term?(term)
        return true if !term.nil? && term != '' && term.to_s.length >= 5

        false
      end

      def add_response_db_filter(term, action_obj, database, schema, table, field)
        return unless valid_term?(term)

        context_filters_by_term[term.to_s].add(ContextFilter.new.for_database(database, schema, table, field, action_obj))
      end

      def add_filter_for_request_parameter(term, rule, parameter_name)
        return unless valid_term?(term)

        context_filters_by_term[term.to_s].add(ContextFilter.new.for_request('form', parameter_name, rule))
      end

      def add_filter_for_header_value(term, rule, header_name)
        return unless valid_term?(term)

        context_filters_by_term[term.to_s].add(ContextFilter.new.for_request('header', header_name, rule))
      end

      def add_filter_for_cookie_value(term, rule, cookie_name)
        return unless valid_term?(term)

        context_filters_by_term[term.to_s].add(ContextFilter.new.for_request('cookie', cookie_name, rule))
      end

      def filter_body!(body)
        dlp_policy = TCellAgent.policy(TCellAgent::PolicyTypes::DATALOSS)
        if dlp_policy && session_id
          session_id_actions = dlp_policy.get_actions_for_session_id
          if session_id_actions
            send_flag = TCellData.filterx(body, session_id_actions.body_event, session_id_actions.body_redact, session_id)
            if send_flag
              TCellAgent.send_event(
                TCellAgent::SensorEvents::DlpEvent.new(
                  route_id,
                  uri,
                  TCellAgent::SensorEvents::DlpEvent::FOUND_IN_BODY,
                  session_id_actions.action_id
                ).for_framework(TCellAgent::SensorEvents::DlpEvent::FRAMEWORK_VARIABLE_SESSION_ID)
              )

            end
          end
        end
        context_filters_by_term.sort_by { |term, _context_filters| -term.length }.each do |term, context_filters|
          replace_filters = (context_filters.select { |context_filter| context_filter.rule.body_redact == true })
          event_filters = (context_filters.select { |context_filter| (context_filter.rule.body_redact != true && context_filter.rule.body_event == true) })
          send_flag = TCellData.filterx(body, !event_filters.empty?, !replace_filters.empty?, term)
          send_flag ||= TCellData.filterx(body, !event_filters.empty?, !replace_filters.empty?, CGI.escapeHTML(term))
          next unless send_flag

          (replace_filters + event_filters).each do |filter|
            base_event = TCellAgent::SensorEvents::DlpEvent.new(
              route_id,
              uri,
              TCellAgent::SensorEvents::DlpEvent::FOUND_IN_BODY,
              filter.rule.action_id
            )
            if filter.type == ContextFilter::DATABASE
              TCellAgent.send_event(
                base_event.for_database(filter.database, filter.schema, filter.table, filter.field)
              )
            elsif filter.type == ContextFilter::REQUEST
              TCellAgent.send_event(
                base_event.for_request(filter.context, filter.parameter)
              )
            end
          end
        end
        body
      end

      def filter_log(log_msg)
        dlp_policy = TCellAgent.policy(TCellAgent::PolicyTypes::DATALOSS)
        if dlp_policy && session_id
          session_id_actions = dlp_policy.get_actions_for_session_id
          if session_id_actions
            send_flag = TCellData.filterx(log_msg, session_id_actions.log_event, session_id_actions.log_redact, session_id)
            if send_flag
              TCellAgent.send_event(
                TCellAgent::SensorEvents::DlpEvent.new(
                  route_id,
                  uri,
                  TCellAgent::SensorEvents::DlpEvent::FOUND_IN_LOG,
                  session_id_actions.action_id
                ).for_framework(TCellAgent::SensorEvents::DlpEvent::FRAMEWORK_VARIABLE_SESSION_ID)
              )
            end
          end
        end
        context_filters_by_term.sort_by { |term, _context_filters| -term.length }.each do |term, context_filters|
          replace_filters = (context_filters.select { |context_filter| context_filter.rule.log_redact == true })
          event_filters = (context_filters.select { |context_filter| (context_filter.rule.log_redact != true && context_filter.rule.log_event == true) })
          send_flag = TCellData.filterx(log_msg, !event_filters.empty?, !replace_filters.empty?, term)
          next unless send_flag

          (replace_filters + event_filters).each do |filter|
            base_event = TCellAgent::SensorEvents::DlpEvent.new(
              route_id,
              uri,
              TCellAgent::SensorEvents::DlpEvent::FOUND_IN_LOG,
              filter.rule.action_id
            )
            if filter.type == ContextFilter::DATABASE
              TCellAgent.send_event(
                base_event.for_database(filter.database, filter.schema, filter.table, filter.field)
              )
            elsif filter.type == ContextFilter::REQUEST
              TCellAgent.send_event(
                base_event.for_request(filter.context, filter.parameter)
              )
            end
          end
        end
        log_msg
      end

      def to_s
        "<#{self.class.name} transaction_id: #{transaction_id} session_id: #{session_id} " \
        "hmac_session_id: #{hmac_session_id} user_id: #{user_id} route_id: #{route_id} " \
        "uri: #{uri} context_filters_by_term: #{context_filters_by_term} " \
        "database_filters: #{database_filters} remote_address: #{remote_address} user_agent: #{user_agent} " \
        "request_method: #{@request_method} path_parameters: #{@path_parameters}>"
      end
    end

    # NOTE: mock for tests
    def self.get_safe_block_logger
      unless defined?(@safe_block_logger)
        @safe_block_logger = TCellAgent::ModuleLogger.new(TCellAgent.logger, name)
      end

      @safe_block_logger
    end

    def self.safe_block(message, &block)
      block.call
    rescue StandardError => e
      logger = get_safe_block_logger
      logger.error("Error #{message} (#{e.class}): #{e.message}")
      logger.exception(e)
    end

    def self.safe_block_no_log(_message, &block)
      block.call
    rescue StandardError
      # do nothing
    end
  end
end
