require 'tcell_agent/policies/policy_types'
require 'tcell_agent/utils/strings'
require 'tcell_agent/configuration'

module TCellAgent
  module Instrumentation
    module Lfi
      extend TCellAgent::ModuleLoggerAccess

      def self.block_file_access?(path, mode)
        TCellAgent::Instrumentation.safe_block('Checking Local Files Policy') do
          if TCellAgent::Utils::Strings.present?(path)
            lfi_policy = TCellAgent.policy(TCellAgent::PolicyTypes::LFI)

            request_env = TCellAgent::Instrumentation::Rails::Middleware::ContextMiddleware::THREADS.fetch(
              Thread.current.object_id, {}
            )

            tcell_context = request_env[TCellAgent::Instrumentation::TCELL_ID]
            return lfi_policy.block_file_access?(path, mode, tcell_context)
          end
        end

        false
      end

      def self.extract_path_mode(*args)
        path = ''
        mode = ''

        TCellAgent::Instrumentation.safe_block('LFI Parsing *args') do
          return ['', ''] if args.nil? || args.empty?

          args_copy = Array.new(args)
          path = args_copy.shift
          mode = args_copy.shift || 'r'

          if path && path.to_s[0] != '|'
            path = File.expand_path(path.to_s)

            mode = if mode && mode.is_a?(Hash)
                     convert_mode(mode[:mode])
                   else
                     convert_mode(mode)
                   end

            [path, mode]
          else
            ['', '']
          end
        end
      end

      def self.extract_path_mode_argf
        path = ''
        mode = 'Read'

        TCellAgent::Instrumentation.safe_block('LFI Parsing ARGF') do
          begin
            return ['', ''] if ARGF.file == $stdin

            if ARGF.eof? && !ARGV.empty?
              argv_copy = Array.new(ARGV)
              path = argv_copy.shift
            else
              path = ARGF.filename
            end

            if path && path.to_s[0] != '|'
              [File.expand_path(path.to_s), mode]
            else
              ['', '']
            end
          rescue Errno::ENOENT
            module_logger.debug('LFI Parsing ARGF: attempted to read a non-existent file')
            ['', '']
          rescue Errno::EISDIR
            module_logger.debug('LFI Parsing ARGF: attempted to read a directory')
            [ARGF.filename, mode]
          end
        end
      end

      def self.convert_mode(mode)
        if mode.is_a? String
          return 'ReadWrite' if mode.include? '+'
          return 'Write' if (mode.include? 'w') || (mode.include? 'a')
        elsif mode.is_a? Numeric
          return 'ReadWrite' if (mode & ::File::RDWR) != 0
          return 'Write' if (mode & ::File::WRONLY) != 0
        end
        'Read'
      end

      def self.raise_if_block(path, mode)
        return unless block_file_access?(path, mode)

        raise IOError, "tCell.io Agent: Attempted access to file #{path} with mode #{mode} denied"
      end

      def self.default_open_handler(args, override_mode = '')
        path, mode = extract_path_mode(*args)

        mode = override_mode unless override_mode.empty?

        raise_if_block(path, mode)
      end

      def self.argf_open_handler
        path, mode = TCellAgent::Instrumentation::Lfi.extract_path_mode_argf

        raise_if_block(path, mode)
      end

      def self.cmdi_open_handler(args, override_mode = '')
        path, mode = extract_path_mode(*args)

        mode = override_mode unless override_mode.empty?

        raise_if_block(path, mode)

        return unless path.empty?

        cmd = TCellAgent::Cmdi.parse_command_from_open(*args)

        TCellAgent::Cmdi.raise_if_block(cmd) if cmd
      end
    end
  end
end
