# See the file "LICENSE" for the full license governing this code.

require 'logger'
require 'tcell_agent/configuration'

module TCellAgent
  class NullLoger < Logger
    def initialize(*args); end

    def add(*args, &block); end
  end

  class ModuleLogger
    def initialize(logger, module_name)
      @logger = logger
      @module_name = module_name
    end

    %i[exception debug info warn error].each do |method_name|
      define_method(method_name) do |msg|
        @logger.send(method_name, @module_name, msg)
      end
    end
  end

  module ModuleLoggerAccess
    def module_logger
      @module_logger ||= ModuleLogger.new(
        TCellAgent.logger, self.class.name
      )
    end
  end

  # NOTE: since the agent waits until native agent
  # is available, this is only used in errors
  # throwned while the agent is instrumenting or starting up
  # so it's ok to send those to STDOUT always
  class RubyLogger
    def initialize
      @logger = Logger.new(STDOUT) # rubocop:disable Style/GlobalStdStream
    end

    def exception(module_name, exception)
      @logger.debug("#{module_name} #{exception.backtrace.join("\n")}")
    end

    %i[debug info warn error].each do |method_name|
      define_method(method_name) do |module_name, msg|
        @logger.send(method_name, "#{module_name} #{msg}")
      end
    end
  end

  class NativeLogger
    def initialize(native_agent)
      @native_agent = native_agent
    end

    def exception(module_name, exception)
      @native_agent.log_message(
        'debug', exception.backtrace.join("\n"), module_name
      )
    end

    %i[debug info warn error].each do |method_name|
      define_method(method_name) do |module_name, msg|
        @native_agent.log_message(
          method_name.to_s, msg, module_name
        )
      end
    end
  end

  @@ruby_logger = RubyLogger.new

  def self.logger
    return @@ruby_logger unless defined?(@native_logger)

    @native_logger
  end

  def self.native_logger=(native_agent)
    @native_logger = NativeLogger.new(native_agent)
  end
end
