module TCellAgent
  require 'base64'
  require 'tcell_agent/agent'

  module DeviseInstrumentation
    module TCellFailureAppRespond
      def respond
        TCellAgent::Instrumentation.safe_block('Devise Failure App Respond') do
          if TCellAgent.configuration.should_intercept_requests?
            tcell_data = request.env[TCellAgent::Instrumentation::TCELL_ID]
            if tcell_data
              # in the case of http auth, user_id is set in
              # Devise::Strategies::Authenticatable.valid_for_http_auth?
              user_id = tcell_data.user_id
              user_id ||= _get_tcell_username

              # in the case of http auth, password is set in
              # Devise::Strategies::Authenticatable.valid_for_http_auth?
              password = tcell_data.password
              password ||= _get_tcell_password

              user_valid = warden_message != :not_found_in_database if defined?(warden_message)

              login_policy = TCellAgent.policy(TCellAgent::PolicyTypes::LOGINFRAUD)
              login_policy.report_login_failure(
                user_id,
                password,
                request.env,
                user_valid,
                tcell_data
              )
            end
          end
        end

        super if defined?(super)
      end

      def _get_tcell_username
        tcell_username = nil
        TCellAgent::Instrumentation.safe_block('Devise Get TCell Username') do
          keys = scope_class.authentication_keys.dup
          user_params = request.POST.fetch('user', {})
          keys.each do |key|
            next_usename = user_params.fetch(key, nil)
            if next_usename
              tcell_username ||= ''
              tcell_username += next_usename
            end
          end
        end
        tcell_username
      end

      def _get_tcell_password
        tcell_password = nil
        TCellAgent::Instrumentation.safe_block('Devise Get TCell Password') do
          user_params = request.POST.fetch('user', {})
          tcell_password = user_params['password']
        end
        tcell_password
      end
    end
  end

  # prepend is ruby 2+ feature
  Devise::FailureApp.send(:prepend, DeviseInstrumentation::TCellFailureAppRespond)

  Devise::Strategies::Authenticatable.class_eval do
    alias_method :tcell_valid_for_http_auth?, :valid_for_http_auth?
    def valid_for_http_auth?
      is_valid = tcell_valid_for_http_auth?

      TCellAgent::Instrumentation.safe_block('Devise set username for http basic auth') do
        tcell_data = request.env[TCellAgent::Instrumentation::TCELL_ID]
        if http_auth_hash && tcell_data
          username = http_auth_hash[http_authentication_key]
          password = http_auth_hash[:password]
          tcell_data.user_id = username if username && !tcell_data.user_id
          tcell_data.password = password if password && !tcell_data.password
        end
      end

      is_valid
    end

    alias_method :tcell_validate, :validate
    def validate(resource, &block)
      is_valid = tcell_validate(resource, &block)
      send_event = is_valid

      # gets the first entry in the current backtrace
      # syntax suggested by rubocop to improve performance
      if caller(1..1).first.include? 'two_factor_authenticatable'
        TCellAgent.logger.debug('Not sending login success event for Devise::Strategies::TwoFactorAuthenticatable since 2fa is unsupported', 'TCellAgent::DeviseInstrumentation')
        send_event = false
      end

      TCellAgent::Instrumentation.safe_block('Devise Authenticatable Validate') do
        if send_event && TCellAgent.configuration.should_intercept_requests?
          username = nil
          (authentication_keys || []).each do |auth_key|
            attr = authentication_hash[auth_key] unless authentication_hash.nil?

            if attr
              username ||= ''
              username += attr
            end
          end

          tcell_data = request.env[TCellAgent::Instrumentation::TCELL_ID]
          return is_valid unless tcell_data

          tcell_data.user_id = username if username && tcell_data.user_id.nil?

          login_policy = TCellAgent.policy(TCellAgent::PolicyTypes::LOGINFRAUD)
          login_policy.report_login_success(
            username,
            request.env,
            tcell_data
          )
        end
      end

      is_valid
    end
  end
end
