require 'tcell_agent/patches'
require 'tcell_agent/rails/routes/grape'
require 'tcell_agent/rails/routes/route_id'
require 'tcell_agent/rails/dlp/process_request'

require 'json'

module TCellAgent
  module Instrumentation
    module Rails
      class TCellRoute
        attr_reader :route_id, :route_path, :route_path_raw, :route_destination

        def initialize(route = nil)
          @route = route

          @route_path_raw = nil
          @route_path = nil
          @route_destination = nil

          return unless route

          @route_path_raw = route.path.spec.to_s

          @route_path = @route_path_raw
          @route_path = @route_path_raw.chomp('(.:format)')

          @route_destination = nil
          @route_destination = JSON.dump(@route.defaults) if @route.defaults
        end

        def report?
          false
        end

        def grape_routes
          []
        end

        def grape_route?
          TCellAgent::Instrumentation.grape_route?(@route)
        end
      end

      class TCellRoute5 < TCellRoute
        def report?
          TCellAgent::Utils::Strings.present?(@route.verb) || grape_route?
        end

        def route_methods
          (@route.verb || '').split('|')
        end

        def grape_routes
          @route.app.app.routes
        end
      end

      class TCellRoute4 < TCellRoute
        METHODS = %w[DELETE GET HEAD OPTIONS PATCH POST PUT TRACE CONNECT].freeze

        def report?
          @route.constraints.key?(:request_method) || grape_route?
        end

        def route_methods
          METHODS.select { |x| @route.verb.match(x) }
        end

        def grape_routes
          if ::Rails::VERSION::MAJOR == 4 && ::Rails::VERSION::MINOR < 2
            @route.app.routes
          else
            @route.app.app.routes
          end
        end
      end

      def self.create_tcell_route(route)
        return TCellRoute5.new(route) if route && ::Rails::VERSION::MAJOR >= 5
        return TCellRoute4.new(route) if route && ::Rails::VERSION::MAJOR < 5

        TCellRoute.new
      end

      def self.instrument_routes
        return unless TCellAgent.configuration.should_instrument?

        return unless ::Rails.application

        ::Rails.application.routes.routes.each do |route|
          instrument_route(route)
        end
      end

      def self.instrument_route(route)
        return unless TCellAgent.configuration.should_instrument?

        tcell_route = create_tcell_route(route)

        return unless tcell_route.report?

        if tcell_route.grape_route?
          TCellAgent::Instrumentation.instrument_grape_api(tcell_route.route_path, tcell_route.grape_routes)

        else
          tcell_route.route_methods.each do |route_method|
            route_id =
              TCellAgent::SensorEvents::Util.calculate_route_id(route_method, tcell_route.route_path_raw)
            TCellAgent.send_event(
              TCellAgent::SensorEvents::AppRoutesSensorEvent.new(
                tcell_route.route_path, route_method, route_id, nil, tcell_route.route_destination
              )
            )
          end
        end
      end

      if ::Rails::VERSION::MAJOR == 3
        ActionDispatch::Routing::RouteSet.class_eval do
          alias_method :tcell_add_route, :add_route
          def add_route(app, conditions = {}, requirements = {}, defaults = {}, name = nil, anchor = true)
            route = tcell_add_route(app, conditions, requirements, defaults, name, anchor)

            TCellAgent::Instrumentation::Rails.instrument_route(route)

            route
          end
        end

        ActiveSupport.on_load(:action_controller) do
          ActionController::Base.class_eval do
            if ::Rails::VERSION::MAJOR == 5
              prepend_around_action :tcell_around_filter_routes
            elsif ::Rails::VERSION::MAJOR < 5
              prepend_around_filter :tcell_around_filter_routes
            end
            def tcell_around_filter_routes
              if TCellAgent.configuration.should_intercept_requests?
                TCellAgent::Instrumentation.safe_block('Determining Rails Route ID') do
                  _match, parameters, route = ::Rails.application.routes.router.recognize(request) { |r, _| r }.first

                  TCellAgent::Instrumentation::RouteId.update_context(env, parameters, route)
                end

                if TCellAgent::Instrumentation::Patches.block?(request)
                  return head(403)
                end

                TCellAgent::DLP.handle_request_dlp_parameters(request)
              end

              yield
            end
          end
        end

      end

      if ::Rails::VERSION::MAJOR == 4
        ActionDispatch::Journey::Routes.class_eval do
          alias_method :tcell_add_route, :add_route
          def add_route(app, path, conditions, defaults, name = nil)
            route = tcell_add_route(app, path, conditions, defaults, name)

            TCellAgent::Instrumentation.safe_block('Reporting new route') do
              TCellAgent::Instrumentation::Rails.instrument_route(route)
            end

            route
          end
        end
      end

      if ::Rails::VERSION::MAJOR >= 5
        ActionDispatch::Journey::Routes.class_eval do
          alias_method :tcell_add_route, :add_route
          def add_route(name, mapping)
            route = tcell_add_route(name, mapping)

            TCellAgent::Instrumentation.safe_block('Reporting new route') do
              TCellAgent::Instrumentation::Rails.instrument_route(route)
            end

            route
          end
        end
      end

      if ::Rails::VERSION::MAJOR >= 5 || (::Rails::VERSION::MAJOR == 4 && ::Rails::VERSION::MINOR >= 2)
        ActionDispatch::Journey::Router.class_eval do
          alias_method :tcell_serve, :serve
          def serve(req)
            if TCellAgent.configuration.should_intercept_requests?
              TCellAgent::Instrumentation.safe_block('Determining Rails Route ID') do
                _match, parameters, route = find_routes(req).first

                TCellAgent::Instrumentation::RouteId.update_context(req.env, parameters, route)
              end

              if TCellAgent::Instrumentation::Patches.block?(req)
                return [403, {}, []]
              end

              TCellAgent::DLP.handle_request_dlp_parameters(req)
            end

            tcell_serve(req)
          end
        end
      end

      if ::Rails::VERSION::MAJOR == 4 && ::Rails::VERSION::MINOR < 2
        require 'action_dispatch/journey/router/utils'

        ActionDispatch::Journey::Router.class_eval do
          alias_method :tcell_call, :call
          def call(env)
            env['PATH_INFO'] = ActionDispatch::Journey::Router::Utils.normalize_path(env['PATH_INFO'])

            if TCellAgent.configuration.should_intercept_requests?
              TCellAgent::Instrumentation.safe_block('Determining Rails Route ID') do
                _match, parameters, route = find_routes(env).first

                TCellAgent::Instrumentation::RouteId.update_context(env, parameters, route)
              end

              tcell_request = Rack::Request.new(env)

              if TCellAgent::Instrumentation::Patches.block?(tcell_request)
                return [403, {}, []]
              end

              TCellAgent::DLP.handle_request_dlp_parameters(tcell_request)
            end

            tcell_call(env)
          end
        end
      end
    end
  end
end
