# frozen_string_literal: true

require 'tcell_agent/version'
require 'tcell_agent/rust/models'

module TCellAgent
  module Rust
    class AgentConfig < Hash
      def initialize(configuration)
        self['agent_type'] = 'Ruby'
        self['agent_version'] = TCellAgent::VERSION
        self['default_cache_dir'] = File.join(Dir.getwd, 'tcell/cache')
        self['default_config_file_dir'] = File.join(Dir.getwd, 'config')
        self['default_log_dir'] = File.join(Dir.getwd, 'tcell/logs')
        self['default_preload_policy_file_dir'] = Dir.getwd

        if defined?(ConfigInitializer)
          overrides = Models.clean_nils(AgentConfigOverrides.new(configuration))
          self['overrides'] = overrides
        else
          self['overrides'] = { 'applications' => [{ :enable_json_body_inspection => true }],
                                'config_file_path' => configuration.get_config_file_path }
        end

        set_agent_details
      end

      def set_agent_details
        framework_details = if defined?(Rails)
                              TCellAgent::Instrumentation::Rails.framework_details
                            else
                              {}
                            end

        self['agent_details'] = { 'language' => 'Ruby',
                                  'language_version' => RUBY_VERSION,
                                  'app_framework' => framework_details['app_framework'],
                                  'app_framework_version' => framework_details['app_framework_version'] }
      end
    end

    class AgentConfigOverrides < Hash
      def initialize(configuration)
        applications = {
          :allow_payloads => configuration.allow_payloads,
          :api_key => configuration.api_key,
          :app_id => configuration.app_id,
          :enable_json_body_inspection => true,
          :hmac_key => configuration.hmac_key,
          :max_header_size => configuration.max_csp_header_bytes,
          :password_hmac_key => configuration.password_hmac_key,
          :reverse_proxy => configuration.reverse_proxy,
          :reverse_proxy_ip_address_header => configuration.reverse_proxy_ip_address_header
        }

        self['api_url'] = configuration.tcell_api_url
        self['applications'] = [Models.clean_nils(applications)]
        self['config_file_path'] = configuration.get_config_file_path
        self['disabled_instrumentation'] = configuration.disabled_instrumentation
        self['enabled'] = configuration.enabled
        self['host_identifier'] = configuration.host_identifier
        self['input_url'] = configuration.tcell_input_url
        self['instrument'] = configuration.instrument
        self['js_agent_api_url'] = configuration.js_agent_api_base_url
        self['js_agent_url'] = configuration.js_agent_url
        self['log_destination'] = configuration.logging_options[:destination]
        self['log_dir'] = configuration.log_dir
        self['log_enabled'] = configuration.logging_options[:enabled]
        self['log_filename'] = configuration.logging_options[:log_filename]
        self['log_level'] = configuration.logging_options[:level]
        self['proxy_url'] = configuration.proxy_url
        self['proxy_username'] = configuration.proxy_username
        self['proxy_password'] = configuration.proxy_password
        self['update_policy'] = configuration.fetch_policies_from_tcell
      end
    end
  end
end
